<?php

namespace Ep\Gath\Plugin;

use Ep\Filter\SanitizeInput;
use Ep\Gath\DataStore\ProductDataStore;
use Ep\Gath\DataStore\VariableDataStore;
use Ep\Gath\DataStore\VariationDataStore;
use Ep\Gath\Gath;
use Ep\Gath\ImageManager;
use Ep\TecDoc\TecDocService;
use Exception;
use Laminas\Log\LoggerInterface;
use Psr\Cache\InvalidArgumentException;
use RuntimeException;
use Solarium\Client;
use Solarium\QueryType\Select\Result\Result;
use WC_Cart;
use WC_Product;
use WP_Post;
use WP_Query;
use WP_Term;

/**
 * Zamjenjuje proizvode onima iz Gatha.
 *
 * @package Ep\Gath\Plugin
 * @author Elektronika Perhoč <web@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/gath
 */
class Products extends AbstractGathPlugin
{
    const ID_OFFSET = 1e7;
    const ID_OFFSET_SIMPLE = self::ID_OFFSET;
    const ID_OFFSET_GROUPED = self::ID_OFFSET_SIMPLE + 1e7;
    const ID_OFFSET_VARIABLE = self::ID_OFFSET_GROUPED + 1e7;
    const ID_OFFSET_VARIATION = self::ID_OFFSET_VARIABLE + 1e7;
    const ID_OFFSET_TIRES = self::ID_OFFSET_VARIATION + 1e7;
    const ID_OFFSET_ATTACHMENT = self::ID_OFFSET_TIRES + 1e7;

    const TERM_OFFSET = 1e7;
    const TERM_OFFSET_BRAND = self::TERM_OFFSET;
    const TERM_OFFSET_CATEGORY = self::TERM_OFFSET_BRAND + 1e7;
    const TERM_OFFSET_ATTRIBUTE = self::TERM_OFFSET_CATEGORY + 1e7;

    /** @var array */
    protected $resultsById = [];
    /** @var array */
    protected $termsById = [];

    /** @var Client */
    protected $search;
    /** @var TecDocService */
    protected $tecdoc;
    /** @var bool */
    protected $initialized = false;

    public function __construct(
        Gath $gath,
        LoggerInterface $logger,
        ImageManager $images,
        Client $search
    )
    {
        parent::__construct($gath, $logger, $images);
        $this->search = $search;
    }

    /**
     * Dodaje akcije i filtere u WordPress.
     *
     * @return void
     */
    public function attach(): void
    {
        if (!$this->initialized) {
            add_filter('get_order_item_metadata', [$this, 'get_order_item_metadata'], 10, 3);
            add_filter('get_post_metadata', [$this, 'get_post_metadata'], 10, 3);
            add_filter('get_term_metadata', [$this, 'get_term_metadata'], 10, 3);
            add_filter('get_the_terms', [$this, 'get_the_terms'], 10, 3);
            add_filter('get_object_terms', [$this, 'get_the_terms'], 10, 4);
            add_filter('posts_pre_query', [$this, 'posts_pre_query'], 10, 2);
            add_filter('update_post_metadata', [$this, 'update_post_metadata'], 10, 2);
            add_filter('woocommerce_add_to_cart_validation', [$this, 'wc_add_to_cart_validation'], 10, 2);
            add_action('init', [$this, 'preload_cart_products'], 20);
            add_action('woocommerce_restore_cart_item', [$this, 'wc_restore_cart_item'], 10, 2);

            add_action('wp_ajax_flatsome_quickview', [$this, 'flatsome_quickview'], 0);
            add_action('wp_ajax_nopriv_flatsome_quickview', [$this, 'flatsome_quickview'], 0);

            add_action('init', [$this, 'init']);
            add_action('init', [$this, 'preload_product_from_get'], 0);

            add_action('plugins_loaded', [$this, 'plugins_loaded']);
            add_filter('woocommerce_redirect_single_search_result', '__return_false');
            add_filter('woocommerce_get_children', [$this, 'wc_get_children'], 10, 2);
            add_filter('woocommerce_product_type_query', [$this, 'wc_product_type_query'], 10, 2);
            add_filter('woocommerce_data_stores', [$this, 'wc_data_stores']);
            add_filter('woocommerce_structured_data_product', [$this, 'wc_structured_data_product'], 19);

            add_filter('ep_gath_products', [$this, 'ep_gath_products']);

            $this->initialized = true;
        }
    }

    public function plugins_loaded()
    {
        if (function_exists('ep_tecdoc_service')) {
            $this->tecdoc = ep_tecdoc_service();
        }
    }

    public function init()
    {
        remove_action('set_object_terms', ['WC_Post_Data', 'set_object_terms']);
        remove_action('transition_post_status', ['WC_Post_Data', 'transition_post_status']);
        remove_action('woocommerce_product_set_visibility', ['WC_Post_Data', 'delete_product_query_transients']);
    }

    /**
     * Ako imamo proizvoda u košarici, potrebno ih je preloadati prije nego što ih WooCommerce
     * pokuša učitati.
     */
    public function preload_cart_products()
    {
        if (!function_exists('WC')) {
            return;
        }

        $session = WC()->session;

        if (is_object($session) && method_exists($session, 'get')) {
            $cart = $session->get('cart');

            foreach ($cart ?? [] as $key => $value) {
                if (isset($value['product_id']) && $value['product_id'] > self::ID_OFFSET) {
                    $this->getProductsByIds([$value['product_id']]);
                }
            }
        }
    }

    public function preload_product_from_get()
    {
        if (($post_id = $_GET['post_id'] ?? null) && $post_id > Products::ID_OFFSET) {
            $this->getProductById(intval($post_id));
        }

        if (isset($_POST['products']) && is_array($_POST['products'])) {
            $this->getProductsByIds(array_keys($_POST['products']));
        }

        if (isset($_GET['add-to-cart']) && is_numeric($_GET['add-to-cart'])) {
            $this->getProductsByIds([$_GET['add-to-cart']]);
        }
    }

    /**
     * @param $metadata
     * @param $object_id
     * @param $meta_key
     *
     * @return mixed
     */
    public function get_order_item_metadata($metadata, $object_id, $meta_key)
    {
        if ('_product_id' === $meta_key) {
            global $wpdb;

            $productId = $wpdb->get_col("SELECT meta_value FROM wp_woocommerce_order_itemmeta WHERE meta_key = '_product_id' AND order_item_id = " . $object_id);

            if (count($productId) > 0) {
                $productId = intval($productId[0]);
                $this->getProductsByIds([$productId]);
            }
        }
        return $metadata;
    }

    /**
     * Ako ova metoda vrati postove, kveri se neće izvršiti.
     *
     * @param $null
     * @param WP_Query $query
     *
     * @return mixed
     *
     * @throws InvalidArgumentException
     */
    public function posts_pre_query($null, WP_Query $query): ?array
    {
        $data = null;

        if ((is_admin() && !is_ajax())) {
            return $null;
        }

        if (in_array($query->query_vars['post_type'], ['product', 'carpart']) && $query->is_search()) {
            return $this->product_search($null, $query);
        }

        // nije kveri za proizvode
        if (
            ($query->query_vars['wc_query'] ?? '') !== 'product_query' &&
            $query->query_vars['post_type'] !== 'product'
        ) {
            return $null;
        }

        // svima će trebat redoslijed
        switch ($query->query_vars['orderby'] ?? '') {
            case 'menu_order title':
            case 'name':
                $sort = Gath::SORT_NAME_ASC;
                break;
            case 'price':
                $sort = $query->query_vars['order'] === 'DESC'
                    ? Gath::SORT_PRICE_DESC
                    : Gath::SORT_PRICE_ASC;
                break;
            case 'date ID':
            default:
                $sort = Gath::SORT_DATE_DESC;
                break;
        }

        // podaci o straničenju također
        $limit = $query->query_vars['posts_per_page'];
        $page = max(1, intval($query->query_vars['paged']));
        $offset = ($page - 1) * $limit;

        $tecdocKeys = $query->query_vars['tecdoc_keys'] ?? [];
        $motoGenIds = $query->query_vars['moto_genids'] ?? [];
        $motoId = absint($query->query_vars['moto_id'] ?? 0);

        $featuredTerm = get_term_by('slug', 'featured', 'product_visibility');

        if (isset($query->query['post__in']) && is_array($query->query['post__in']) && count($query->query['post__in']) > 0) {
            $ids = [];

            foreach ($query->query['post__in'] as $id) {
                if ($id > self::ID_OFFSET) {
                    $ids[] = $id;
                }
            }

            if (count($ids) > 0) {
                $products = $this->getProductsByIds($ids, $limit, $offset);

                if (count($products) > 0) {
                    $query->set('no_found_rows', false);
                    $query->found_posts = count($ids);
                    $query->max_num_pages = ceil($query->found_posts / $limit);
                    $query->thumbnails_cached = true;
                }

                return $products;
            }
        } elseif ('product' === ($query->query_vars['post_type'] ?? '') && isset($query->query_vars['product_id'])) {
            $productId = $query->query_vars['product_id'];

            return $this->getProductsByIds([$productId], $limit, $offset);
            // da li tražimo istaknute proizvode
        } elseif (
            isset($query->tax_query->queried_terms['product_visibility']) &&
            (in_array('featured', $query->tax_query->queried_terms['product_visibility']['terms']) ||
                in_array($featuredTerm->term_id, $query->tax_query->queried_terms['product_visibility']['terms']))
        ) {
            $data = $this->gath->getFeaturedProducts($sort, $limit, $offset);
        } elseif ($motoId > 0 && count($motoGenIds) > 0) {
            $data = $this->gath->getProductsByMotorcycle($motoId, $motoGenIds, $sort, $limit, $offset);
            // ako je ovo TecDoc kategorija
        } elseif (count($tecdocKeys) > 0) {
            $tecDocPairs = [];

            foreach ($tecdocKeys as $key) {
                [$brandNo, $articleNo] = explode('|', $key);

                $tecDocPairs[] = [
                    'BrandNo' => intval($brandNo),
                    'ArticleNo' => $articleNo,
                ];
            }

            $data = $this->gath->getProductsByTecDocPairs($tecDocPairs, $sort, $limit, $offset);
        } elseif (($query->query_vars['taxonomy'] ?? '') === 'product_cat') {
            $category = get_term_by('slug', $query->query_vars['term'], 'product_cat');

            $config = $this->gath->getConfig();

            if (in_array($category->slug, $config['no_products_for_cats'] ?? [])) {
                return [];
            }

            $isTireCategory = (bool)get_term_meta($category->term_id, 'IsTyreCategory', true);

            if ($isTireCategory) {
                $allowedKeys = [
                    'TireWidth' => 'width',
                    'TireHeight' => 'height',
                    'TireDiameter' => 'diameter',
                    'ArtApply' => 'type',
                    'VehicleType' => 'vehicle',
                    'Season' => 'season',
                    'SpeedIdx' => 'speedIndex',
                    'BrandName' => 'brand',
                    'LoadIndex1' => 'loadIndex'
                ];

                $filters = [];
                foreach ($allowedKeys as $gathKey => $ourKey) {
                    if (isset($_GET[$ourKey]) && !empty($_GET[$ourKey])) {
                        $filters[$gathKey] = $_GET[$ourKey];
                    } else if (isset($_GET[$gathKey]) && !empty($_GET[$gathKey])) {
                        $filters[$gathKey] = $_GET[$gathKey];
                    }
                }

                if (!empty($filters)) {
                    $data = $this->gath->getTires($filters, $sort, $limit, $offset);
                }
            } else {
                if (!$this->gath->usesWooCategories()) {
                    $categoryId = intval($category->term_id - Products::TERM_OFFSET_CATEGORY);
                } else {
                    $categoryId = $category->term_id;
                }

                $brands = $_GET['brand'] ?? [];
                $taxonomies = wc_get_attribute_taxonomies();
                $attributes = [];

                foreach ($_GET as $name => $value) {
                    if (preg_match('/^pa_(.+)$/', $name, $matches) && is_array($value) && !empty($value)) {
                        $attributeId = intval($matches[1]);
                        $attributes[$attributeId] = array_map('stripslashes', array_map('rawurldecode', $value));
                    }
                }

                if (!empty($attributes)) {
                    $data = $this->gath->getProductsByAttributes($categoryId, $attributes, $brands, $sort, $limit, $offset);
                } else {
                    $data = $this->gath->getProductsByCategory($categoryId, $sort, $limit, $offset, $brands);
                }
            }
            // proizvod u glavnom kveriju
        } elseif (preg_match('/-(\d{8})$/', $query->query_vars['name'], $matches)) {
            $productId = intval($matches[1]);
            return $this->getProductsByIds([$productId]);
        }


        if (isset($data) && count($data['Results']) > 0) {
            $query->set('no_found_rows', false);
            $query->found_posts = $data['RecordCount'];
            $query->max_num_pages = ceil($query->found_posts / $limit);
            $query->thumbnails_cached = true;

            return $this->loadPosts($data['Results']);
        }

        return null;
    }

    /**
     * @param $null
     * @param WP_Query $query
     *
     * @return array
     *
     * @throws InvalidArgumentException
     */
    public function product_search($null, WP_Query $query): ?array
    {
        if ((is_admin() && !is_ajax()) || !in_array($query->query_vars['post_type'], ['product', 'carpart']) || !$query->is_search()) {
            return $null;
        }

        $tecdoc_keys = $query->get('tecdoc_keys');
        $products = [];

        if (!is_array($tecdoc_keys)) {
            $tecdoc_keys = [];
        }

        $q = $query->get('s');
        $tecdocExclusive = $this->gath->getConfig()['tecdoc_exclusive'] ?? false;

        if (strlen($q) > 0) {
            if (!$tecdocExclusive || ($_GET['tecdoc'] ?? '') === '') {
                try {
                    $queryString = sprintf('name:"%1$s" OR sku:"%2$s" OR attribute:"%1$s" OR katbroj:"%1$s" OR katbroj:"%2$s"', $q, strtoupper($q));
                    $results = $this->getSearchResults($queryString);

                    if ($results->count() === 0) {
                        $queryString = sprintf('name:%1$s~1 OR sku:%1$s OR attribute:%1$s', $q);
                        $results = $this->getSearchResults($queryString);
                    }
                } catch (Exception $e) {
                    $results = [];
                }

                $productIds = [];
                $genericArticleIds = [];

                foreach ($results as $result) {
                    if (in_array('product', $result->type)) {
                        if ($result->id > self::ID_OFFSET_TIRES) {
                            continue;
                        }

                        if (is_array($result->parentid) && count($result->parentid) > 0) {
                            $productIds[] = self::ID_OFFSET_GROUPED + $result->parentid[0];
                        } else {
                            $productIds[] = absint($result->id);
                        }
                    } else if (in_array('category', $result->type)) {
                        $multiIds = $result->tecdocmultiid[0] ?? '';

                        if (strlen($multiIds) > 0) {
                            $genericArticleIds = array_merge(
                                $genericArticleIds,
                                array_map('intval', explode(';', $result->tecdocmultiid[0] ?? ''))
                            );
                        }
                    }
                }

                $productIds = array_unique($productIds);

                if (count($genericArticleIds) > 0 && $this->tecdoc) {
                    $genericArticleIds = array_unique($genericArticleIds);
                    $articles = $this->tecdoc->getArticlesByGenericIds($genericArticleIds);

                    foreach ($articles as $article) {
                        $tecdoc_keys[] = $article['key'];
                    }
                }

                if (count($productIds) > 0) {
                    $products = array_merge($products, $this->getProductsByIds($productIds));
                }
            }

            if (count($tecdoc_keys) > 0 && (!$tecdocExclusive || ($_GET['tecdoc'] ?? '') === 'exclusive')) {
                $tecdocProducts = $this->getProductsByTecDocKeys($tecdoc_keys, 1, 1000);
                $products = array_merge($products, $tecdocProducts);
            }
        }

        $orderby = $query->query_vars['orderby'];

        switch ($orderby) {
            case 'name':
                usort($products, [$this, 'sort_by_name']);
                break;
            case 'price':
                usort($products, [$this, 'sort_by_price']);
                break;
            case 'price-desc':
                usort($products, [$this, 'sort_by_price']);
                $products = array_reverse($products);
                break;
        }

        $query->set('no_found_rows', false);
        $query->found_posts = count($products);
        $query->max_num_pages = 1;
        $query->thumbnails_cached = true;
        $query->set('posts_per_page', -1);

        wc_set_loop_prop('total', count($products));
        return count($products) > 0 ? $products : $null;
    }

    public function sort_by_name(WP_Post $a, WP_Post $b): int
    {
        return $a->post_title < $b->post_title ? -1 : ($a->post_title > $b->post_title ? 1 : 0);
    }

    public function sort_by_price(WP_Post $a, WP_Post $b): int
    {
        $a = wc_get_product($a->ID);
        $b = wc_get_product($b->ID);

        return $a->get_price() < $b->get_price() ? -1 : ($a->get_price() > $b->get_price() ? 1 : 0);
    }

    /**
     * @param $productId
     *
     * @return array
     */
    public function getRawResult($productId): array
    {
        return $this->resultsById[$productId] ?? [];
    }

    /**
     * @param $queryString
     *
     * @return Result
     */
    protected function getSearchResults($queryString): Result
    {
        $searchQuery = $this->search->createSelect();
        $searchQuery->setQuery($queryString);
        $searchQuery->setRows(10000);

        return $this->search->select($searchQuery);
    }

    /**
     * Vraća taksonomske pojmove vezane uz proizvod.
     *
     * @param $terms
     * @param $object_id
     * @param $taxonomy
     *
     * @return array
     */
    public function get_the_terms($terms, $object_id, $taxonomy, $args = [])
    {
        if(is_array($taxonomy) && count($taxonomy) > 0) {
            $taxonomy = $taxonomy[0];
        }

        if (!in_array($taxonomy, [
            'featured_item_category',
            'featured_item_tag',
            'product_tag',
            'product_cat',
            'product_brand',
            'product_shipping_class',
            'product_type',
            'product_visibility',
        ])) {
            return $terms;
        }

        if (!is_array($object_id)) {
            $object_id = [$object_id];
        }

        foreach ($object_id as $id) {
            if ($id > self::ID_OFFSET && isset($this->resultsById[$id])) {
                $result = $this->resultsById[$id];

                switch ($taxonomy) {
                    case 'featured_item_category':
                    case 'featured_item_tag':
                    case 'product_tag':
                        break;
                    case 'product_cat':
                        $genericArticleId = get_post_meta($id, '_generic_article_id', true);

                        if ($genericArticleId) {
                            $found = get_terms([
                                'taxonomy' => 'product_cat',
                                'genericArticleId' => $genericArticleId,
                            ]);
                        } else {
                            if ($this->gath->usesWooCategories()) {
                                if (isset($result['woocategories']) && is_array($result['woocategories']) && !empty($result['woocategories'])) {
                                    $found = array_filter(array_map(function (int $categoryId) {
                                        return get_term($categoryId, 'product_cat');
                                    }, $result['woocategories']), function ($cat) { return $cat instanceof WP_Term; });
                                }
                            } else {
                                $found = array_filter(array_map(function (int $categoryId) {
                                    $categoryId = intval($categoryId + Products::TERM_OFFSET_CATEGORY);
                                    return get_term($categoryId, 'product_cat');
                                }, $result['categories'] ?? []), function ($cat) {
                                    return $cat instanceof WP_Term && $cat->name !== null;
                                });
                            }
                        }
                        break;
                    case 'product_brand':
                        if (isset($result['Brands']) && is_array($result['Brands']) && count($result['Brands']) === 1) {
                            $found = [$this->loadBrand($result['Brands'][0])];
                        } else if (isset($result['brandid']) && $result['brandid'] > 0) {
                            $brand = get_term($result['brandid'], 'product_brand');
                            if ($brand instanceof WP_Term) {
                                $found = [$brand];
                            }
                        }
                        break;
                    case 'product_shipping_class':
                        $found = ($result['shippingclassid'] ?? 0) > 0
                            ? [get_term($result['shippingclassid'])]
                            : null;
                        break;
                    case 'product_type':
                        $found = [get_term_by('slug', ($result['type'] ?? 'simple'), 'product_type')];
                        break;
                    case 'product_visibility':
                        $found = isset($result['featured'])
                            ? [get_term_by('slug', 'featured', 'product_visibility')]
                            : [];
                        break;
                }

                if (isset($found)) {
                    if (($args['fields'] ?? '') === 'ids') {
                        $terms = array_map(function (WP_Term $term) {
                            return $term->term_id;
                        }, $found);
                    } else if (($args['fields'] ?? '') === 'tt_ids') {
                        $terms = array_map(function (WP_Term $term) {
                            return $term->term_id;
                        }, $found);
                    } else {
                        $terms = $found;
                    }
                }
            }
        }

        return $terms;
    }

    /**
     * @param $null
     * @param $object_id
     * @param $meta_key
     *
     * @return mixed
     */
    public function get_post_metadata($null, $object_id, $meta_key)
    {
        if ($object_id > self::ID_OFFSET && isset($this->resultsById[$object_id])) {
            $result = $this->resultsById[$object_id];

            if ($result['type'] === 'tire' && !empty($result['image']) && preg_match('#^(https?:)//#', $result['image'])) {
                $images = [
                    $this->images->getImageId('external-' . $result['image'])
                ];
            } else {
                if (isset($result['Images']) && is_array($result['Images']) && !empty($result['Images'])) {
                    $images = $result['Images'];
                } else if (isset($result['image']) && !empty($result['image'])) {
                    $images = [$result['image']];
                } else {
                    $images = [];
                }

                $images = array_map(function ($image) {
                    $image = is_array($image) ? ($image['url'] ?? '') : $image;
                    return $image ? $this->images->getImageId($image) : null;
                }, $images);

                $images = array_filter($images);
            }

            switch ($meta_key) {
                case '_automodel':
                    return $result['automodel'] ?? '';
                case '_automotor':
                    return $result['automotor'] ?? '';
                case '_brand':
                    return [$result['Brands'][0] ?? []];
                case '_backorders':
                case '_downloadable':
                case '_sold_individualy':
                case '_virtual':
                    return 'no';
                case '_children':
                    $childrenIds = [];
                    $children = $this->loadPosts($result['GroupedChildren'] ?? []);

                    foreach ($children as $child) {
                        $childrenIds[] = $child->ID;
                    }

                    return [$childrenIds];
                case '_product_image_gallery':
                    if (count($images) > 1) {
                        array_shift($images);
                        return implode(',', $images);
                    }
                    return $null;
                case '_sku':
                    return $result['sku'] ?? $result['SupplierSku'] ?? '';
                case '_tax_class':
                    return $result['tax_class'] ?? ($result['tax_classHR'] ?? '');
                case '_tax_status':
                    return 'taxable';
                case '_tecdoc_key':
                case '_tecDoc_key':
                    return (isset($result['wsbrandno']) && isset($result['wsartno']))
                        ? sprintf('%s|%s', $result['wsbrandno'], $result['wsartno'])
                        : '';
                case '_thumbnail_id':
                    return $images[0] ?? $null;
                case '_crosssell_ids':
                case '_default_attributes':
                case '_download_expiry':
                case '_download_limit':
                case '_height':
                case '_length':
                case '_purchase_note':
                case '_sale_price_dates_from':
                case '_sale_price_dates_to':
                case '_upsell_ids':
                case '_weight':
                case '_width':
                case 'total_sales':
                    return '';
            }
        }

        return $null;
    }

    public function get_term_metadata($null, $object_id, $meta_key)
    {
        if (!isset($this->termsById[$object_id])) {
            return $null;
        }

        $term = $this->termsById[$object_id];

        switch ($meta_key) {
            case 'thumbnail_id':
                if (isset($term['image']) && !empty($term['image'])) {
                    return $this->images->getImageId($term['image']);
                }

                return '';
            default:
                return '';
        }
    }

    /**
     * Ne dopuštamo ažuriranje meta podataka za postove iz Gatha.
     *
     * @param $null
     * @param $object_id
     *
     * @return bool
     */
    public function update_post_metadata($null, $object_id): ?bool
    {
        if ($object_id > self::ID_OFFSET) {
            return true;
        }

        return $null;
    }

    public function ep_gath_products(): Products
    {
        return $this;
    }

    /**
     * Prije nego što flatsome renda svoj quickview.
     */
    public function flatsome_quickview()
    {
        $this->getProductsByIds([intval($_POST['product'])]);
    }

    /**
     * Prije nego što dodajemo proizvod u košaricu.
     *
     * @param $valid
     * @param $product_id
     *
     * @return mixed
     */
    public function wc_add_to_cart_validation($valid, $product_id)
    {
        $this->getProductsByIds([$product_id]);
        return $valid;
    }

    /**
     * @param array $stores
     *
     * @return array
     */
    public function wc_data_stores(array $stores): array
    {
        $stores['product'] = ProductDataStore::class;
        $stores['product-variable'] = VariableDataStore::class;
        $stores['product-variation'] = VariationDataStore::class;
        return $stores;
    }

    /**
     * @param $children
     * @param WC_Product $product
     *
     * @return array
     */
    public function wc_get_children($children, WC_Product $product): array
    {
        if (isset($this->resultsById[$product->get_id()])) {
            $result = $this->resultsById[$product->get_id()];
            $childrenIds = [];

            if (isset($result['GroupedChildren']) && is_array($result['GroupedChildren'])) {
                foreach ($result['GroupedChildren'] as $child) {
                    $childrenIds[] = $this->generateProductId($child);
                }
            }

            return $childrenIds;
        }
        return $children;
    }

    /**
     * @param $false
     * @param $product_id
     *
     * @return string
     */
    public function wc_product_type_query($false, $product_id): string
    {
        if ($product_id > self::ID_OFFSET_TIRES) {
            return 'tire';
        } else if ($product_id > self::ID_OFFSET_VARIATION) {
            return 'variation';
        } else if ($product_id > self::ID_OFFSET_VARIABLE) {
            return 'variable';
        } else if ($product_id > self::ID_OFFSET_GROUPED) {
            return 'grouped';
        } else if ($product_id > self::ID_OFFSET_SIMPLE) {
            return 'simple';
        }

        return $false;
    }

    /**
     * Kada restoramo proizvod u košaricu, napunimo ga nazad proizvodom iz Gatha.
     *
     * @param string $cart_item_key
     * @param WC_Cart $cart
     */
    public function wc_restore_cart_item(string $cart_item_key, WC_Cart $cart)
    {
        $cartItem = $cart->cart_contents[$cart_item_key];

        $productId = absint($cartItem['variation_id'] ?: $cartItem['product_id']);

        if ($productId > 0) {
            $this->getProductById($productId);
            $cart->cart_contents[$cart_item_key]['data'] = wc_get_product($productId);
        }
    }

    public function wc_structured_data_product($markup)
    {

        return $markup;
    }

    /**
     * @param int $categoryId
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return array|WP_Post
     */
    public function getProductsByCategory(int $categoryId, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $data = $this->gath->getProductsByCategory($categoryId, $orderBy, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * Generira ID koji koristimo unutar WordPress-a.
     *
     * @param array $product
     *
     * @return int
     */
    public function generateProductId(array $product): int
    {
        switch ($product['type'] ?? $product['Type'] ?? 'simple') {
            case 'tire':
                return self::ID_OFFSET_TIRES + $product['id'];
            case 'variation':
                return self::ID_OFFSET_VARIATION + $product['id'];
            case 'variable':
                return self::ID_OFFSET_VARIABLE + $product['id'];
            case 'grouped':
                return self::ID_OFFSET_GROUPED + $product['id'];
            case 'simple':
            default:
                return self::ID_OFFSET_SIMPLE + $product['id'];
        }
    }

    /**
     * @param int $id
     *
     * @return WP_Post
     */
    public function getProductById(int $id): WP_Post
    {
        $products = $this->getProductsByIds([$id]);

        if (0 === count($products)) {
            throw new RuntimeException(sprintf("Proizvod s id-em %d ne postoji u Gathu.", $id));
        }

        return $products[0];
    }

    /**
     * @param array $ids
     *
     * @param int $limit
     * @param int $offset
     * @return WP_Post[]
     */
    public function getProductsByIds(array $ids, int $limit = 1000, int $offset = 0): array
    {
        $data = $this->gath->getProductsByIds($ids, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $skus
     *
     * @return WP_Post[]
     */
    public function getProductsBySkus(array $skus): array
    {
        $data = $this->gath->getProductsByskus($skus);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $keys
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return WP_Post[]
     */
    public function getProductsByTecDocKeys(array $keys, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $tecDocPairs = [];

        foreach ($keys as $key) {
            [$brandNo, $articleNo] = explode('|', $key);

            $tecDocPairs[] = [
                'BrandNo' => intval($brandNo),
                'ArticleNo' => $articleNo,
            ];
        }

        return $this->getProductsByTecDocPairs($tecDocPairs, $orderBy, $limit, $offset);
    }

    /**
     * @param array $pairs
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return WP_Post[]
     */
    public function getProductsByTecDocPairs(array $pairs, int $orderBy = Gath::SORT_DATE_DESC, int $limit = 10, $offset = 0): array
    {
        $data = $this->gath->getProductsByTecDocPairs($pairs, $orderBy, $limit, $offset);
        return $this->loadPosts($data['Results']);
    }

    /**
     * @param array $brand
     *
     * @return WP_Term
     */
    protected function loadBrand(array $brand): WP_Term
    {
        $brandId = $brand['brandid'] + self::TERM_OFFSET_BRAND;

        $this->termsById[$brandId] = $brand;

        return new WP_Term((object)[
            'term_id' => $brandId,
            'name' => $brand['name'],
        ]);
    }

    /**
     * @param array $results
     *
     * @return array|WP_Post[]
     */
    protected function loadPosts(array $results): array
    {
        $posts = [];

        foreach ($results as $result) {
            $postId = $this->generateProductId($result);
            $this->resultsById[$postId] = $result;

            $postDate = preg_replace('/\.\d+$/', '', $result['date_modified'] ?? date('Y-m-d H:i:s'));
            $postTitle = $result['name'] ?? '';

            if ($result['type'] === 'tire') {
                $postTitle = sprintf(
                    '%s %s %d/%dR%d %d%s',
                    $result['brand_name'],
                    $result['name'],
                    $result['tire_width'],
                    $result['tire_height'],
                    $result['tire_diameter'],
                    $result['load_index1'],
                    $result['speed_idx']
                );
            }

            $filter = new SanitizeInput();
            $postName = sprintf('%s-%d', $filter->filter($result['name'] ?? ''), $postId);
            $description = $this->gath->areDescriptionsFiltered()
                ? preg_replace('/<([a-z]+)[^>]+>/', '<\1>', $result['description'] ?? '')
                : $result['description'];
            $parentId = $result['type'] === 'variation' ? intval(self::ID_OFFSET_VARIABLE + $result['parentid']) : 0;

            if ($result['type'] === 'variation' && !get_post($parentId)) {
                $this->getProductsByIds([$parentId]);
            }

            $post = get_post(new WP_Post((object)[
                'ID' => $postId,
                'post_author' => 1,
                'post_date' => $postDate,
                'post_date_gmt' => $postDate,
                'post_content' => $description,
                'post_title' => $postTitle,
                'post_excerpt' => $this->gath->getShowShortDescription() ? ($result['short_description'] ?? '') : '',
                'post_status' => 'publish',
                'comment_status' => 'closed',
                'ping_status' => 'closed',
                'post_password' => '',
                'post_name' => $postName,
                'to_ping' => '',
                'pinged' => '',
                'post_modified' => $postDate,
                'post_modified_gmt' => $postDate,
                'post_content_filtered' => '',
                'post_parent' => $parentId,
                'guid' => sprintf('%s?p=%d', get_option('siteurl'), $postId),
                'post_type' => $result['type'] === 'variation' ? 'product_variation' : 'product',

                'filter' => 'raw',
            ]));

            wp_cache_add($postId, $post, 'posts');

            if (isset($result['GroupedChildren']) && is_array($result['GroupedChildren'])) {
                $this->loadPosts($result['GroupedChildren']);
            }

            $posts[] = $post;
        }

        global $wp_query;
        return apply_filters('the_posts', $posts, $wp_query);
    }

    protected function getExternalImageUrl($image): string
    {
        $urlParts = parse_url($image);

        // Bail if url not valid
        if ( !filter_var($urlParts['scheme'] . '://' . $urlParts['host'], FILTER_VALIDATE_URL) )
            return '';

        // If scheme is https, we can handle it by default
        if ( $urlParts['scheme'] === 'https')
            return $image;

        // Return string in this format:
        // external-host.tld/[path/]image.jpg
        return 'external-' . $urlParts['host'] . $urlParts['path'];
    }
}
